from inspect import signature, Parameter
from itertools import chain, product, permutations
from typing import Callable, get_type_hints, Iterable, Tuple
from warnings import warn

from dyndis.type_keys.type_key import TypeKey, type_keys, CoreTypeKey, Self
from dyndis.util import SubPriority

try:
    from typing import Literal
except ImportError:
    Literal = None


def includes_keys(lhs: Tuple[TypeKey, ...], rhs: Tuple[TypeKey, ...]):
    """
    :return: whether each typekey in lhs is less than or equal to at least one type key in rhs
    """
    for left in lhs:
        for right in rhs:
            try:
                if left <= right:
                    break
            except TypeError:
                continue
        else:
            return False
    return True


_missing = object()


class Candidate:
    """
    A class representing a specific implementation for a multi-dispatch
    """

    def __init__(self, types: Iterable[CoreTypeKey], func: Callable, priority):
        """
        :param types: the types for the conditions
        :param func: the function of the implementation
        :param priority: the priority of the candidate over other candidates (higher is tried first)
        """
        self.types = tuple(types)
        self.func = func
        self.priority = priority
        self.__name__ = getattr(func, '__name__', None)
        self.__doc__ = getattr(func, '__doc__', None)

    @classmethod
    def from_func(cls, priority, func, fallback_type_hint=_missing, self_type=_missing):
        """
        create a list of candidates from function using the function's type hints. ignores all parameters with default
        values, as well as variadic parameters or keyword-only parameters

        :param priority: the priority of the candidate
        :param func: the function to use
        :param fallback_type_hint: the default type hint to use for parameters with missing hints
         this function
        :param self_type: the type to put in place of dyndis.Self (if found)

        :return: a list of candidates generated from the function
        """

        type_hints = get_type_hints(func)

        params = signature(func).parameters
        type_iters = []
        super_type_iters = [type_iters]
        p: Parameter
        for p in params.values():
            if p.kind not in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD):
                break
            t = type_hints.get(p.name, fallback_type_hint)
            if t is _missing:
                if p.default is p.empty:
                    break
                raise KeyError(p.name)

            i = type_keys(t)
            if p.default is not p.empty:
                default_type = type(p.default)
                if default_type is not object:
                    default_type_keys = type_keys(default_type)
                    if not includes_keys(default_type_keys, i):
                        if p.default is None:
                            i = {*i, *default_type_keys}
                        else:
                            warn(f'default value for parameter {p.name} is not included in its annotations',
                                 stacklevel=3)
                super_type_iters.append(list(type_iters))
            type_iters.append(i)

        type_lists = chain.from_iterable(product(*ti) for ti in super_type_iters)
        ret = []
        if self_type is not _missing:
            stk = type_keys(self_type)
            if len(stk) != 1 or not isinstance(stk[0], CoreTypeKey):
                raise TypeError('self_type must evaluate to a single core type key')
            self_type_key = stk[0]
        else:
            self_type_key = None

        for types in type_lists:
            types = list(types)

            encountered_tvars = set()
            tk: TypeKey
            for i, tk in enumerate(types):
                if tk is Self:
                    if self_type_key is None:
                        raise Exception('cannot use dyndis.Self outside of implementors')
                    tk = types[i] = self_type_key
                if not isinstance(tk, CoreTypeKey):
                    raise Exception('split returned a non-core type key')
                tk.introduce(encountered_tvars)
            ret.append(cls(
                types, func,
                priority
            ))
        return ret

    def __str__(self):
        def type_name(t):
            if isinstance(t, type):
                return t.__name__
            return str(t)

        return (self.__name__ or 'unnamed candidate') + '<' + ', '.join(type_name(n) for n in self.types) + '>'

    def permutations(self):
        """
        create a list of candidates from a single candidate, such that they all permutations
        of the candidate's types will be accepted. Useful for symmetric functions.

        :return: a list of equivalent candidates, differing only by the type order
        """
        if not self.types:
            raise ValueError("can't get permutations of a 0-parameter candidate")
        ret = []
        name = getattr(self.func, '__name__', None)
        call_args = ', '.join('_' + str(i) for i in range(len(self.types)))
        seen = set()
        glob = {'__original__': self.func}
        first = True
        for perm in permutations(range(len(self.types))):
            if first:
                func = self.func
                t = self.types
                seen.add(t)
                priority = self.priority
                first = False
            else:
                t = tuple(self.types[i] for i in perm)
                if t in seen:
                    continue
                seen.add(t)

                args = ", ".join('_' + str(i) for i in perm)
                ns = {}
                exec(
                    f"def func({args}): return __original__({call_args})",
                    glob, ns
                )
                func = ns['func']
                if name:
                    func.__name__ = name
                priority = SubPriority(self.priority, -1)
            ret.append(
                Candidate(t, func, priority)
            )
        return ret

    def __lt__(self, other):
        if self.types == other.types or len(self.types) != len(other.types):
            return False
        ret = False
        for left, right in zip(self.types, other.types):
            if ret:
                try:
                    if not left <= right:
                        return False
                except TypeError:
                    return False
            else:
                try:
                    if left < right:
                        ret = True
                    elif not (left <= right):
                        return False
                except TypeError:
                    return False
        return ret

    def __le__(self, other):
        if len(self.types) != len(other.types):
            return False
        if self.types == other.types and self.priority == other.priority:
            return True
        return self < other



from functools import partial

from dyndis.exceptions import NoCandidateError


class MultiDispatchDelegate:
    """
    A base class for multidispatch delegates
    """

    def __init__(self, md):
        self.md = md

    def __set_name__(self, owner, name):
        if not self.md.__name__:
            self.md.__name__ = name


class MultiDispatchOp(MultiDispatchDelegate):
    """
    An operator adapter for a MultiDispatch
    """

    def __get__(self, instance, owner):
        if instance:
            return partial(self.__call__, instance)
        return self

    def __call__(self, *args, **kwargs):
        return self.md.get(args, kwargs, default=NotImplemented)


_no_answer = object()


class MultiDispatchRaisingDelegate(MultiDispatchDelegate):
    def __call__(self, *args, **kwargs):
        ret = self.md.get(args, kwargs, default=_no_answer)
        if ret is _no_answer:
            raise NoCandidateError(args)
        return ret


class MultiDispatchMethod(MultiDispatchRaisingDelegate):
    """
    A method adapter for a MultiDispatch
    """

    def __get__(self, instance, owner):
        if instance:
            return partial(self.__call__, instance)
        return self


class MultiDispatchClassMethod(MultiDispatchRaisingDelegate):
    """
    A static method adapter for a MultiDispatch
    """

    def __get__(self, instance, owner):
        return partial(self.__call__, owner)


class MultiDispatchStaticMethod(MultiDispatchRaisingDelegate):
    """
    A static method adapter for a MultiDispatch
    """

    def __get__(self, instance, owner):
        return self



from __future__ import annotations

from functools import partial
from itertools import chain
from typing import Dict, Tuple, List, Callable, Union, Iterable, Optional, Iterator

from dyndis.candidate import Candidate
from dyndis.descriptors import MultiDispatchOp, MultiDispatchMethod, MultiDispatchStaticMethod, MultiDispatchClassMethod
from dyndis.exceptions import NoCandidateError, AmbiguityError
from dyndis.implementor import Implementor
from dyndis.topological_set import TopologicalSet, CandidateSet
from dyndis.type_keys.type_key import MatchException
from dyndis.util import RawReturnValue

RawNotImplemented = RawReturnValue(NotImplemented)

class CachedSearch:
    """
    A cached search for candidates of a specific type tuple
    """

    def __init__(self, owner: MultiDispatch, key: Tuple[type, ...]):
        """
        :param owner: the owning multidispatch
        :param key: the type tuple to use
        """
        self.query = key

        self.layer_iter = iter(owner.candidate_set.layers)

        self.sorted = []

    def advance(self):
        """
        advance the search by one layer, and cache the results
        :return: the new candidates added to the sorted list
        """
        layer: Optional[CandidateSet] = next(self.layer_iter, None)
        if layer is None:
            self.layer_iter = None
            return None

        try:
            sub_layers = layer.by_sublayer(self.query)
        except MatchException:
            # ensure the error will be raised again if called
            self.layer_iter = chain([layer], self.layer_iter)
            raise

        sl = [[c.inner for c in s] for s in sub_layers]
        self.sorted.extend(sl)
        return sl

    def __iter__(self):
        yield from self.sorted
        while self.layer_iter:
            next_layers = self.advance()
            if next_layers is not None:
                yield from next_layers


EMPTY = object()


class MultiDispatch:
    """
    The central class, a callable that can delegate to multiple candidates depending on the types of parameters
    """

    def __init__(self, name: str = None, doc: str = None):
        """
        :param name: an optional name for the callable
        :param doc: an optional doc for the callable
        """
        self.__name__ = name
        self.__doc__ = doc

        # self.candidate_trie: CandTrie = RankedChildrenTrie()
        self.candidate_set: TopologicalSet[Candidate] = TopologicalSet()
        self.cache: Dict[int, Dict[Tuple[type, ...], CachedSearch]] = {}

    def _clean_cache(self, sizes: Iterable[int]):
        """
        clear the candidate cache for all type tuples of the sizes specified

        :param sizes: the sizes for which to clear to cache
        """
        for size in sizes:
            self.cache.pop(size, None)

    def _add_candidate(self, candidate: Candidate, clean_cache=True):
        """
        Add a single candidate to the multidispatch. If the multidispatch has no set name or doc, the name or doc of
        the candidate will be used (if available)

        :param candidate: the candidate to add
        :param clean_cache: whether to clean the relevant cache
        """
        if not self.candidate_set.add(candidate):
            raise ValueError(f'A candidate of equal types ({candidate.types})'
                             f' and priority ({candidate.priority}) exists')

        if not self.__name__:
            self.__name__ = candidate.__name__
        if not self.__doc__:
            self.__doc__ = candidate.__doc__
        if clean_cache:
            self._clean_cache((len(candidate.types),))

    def add_candidates(self, candidates: Iterable[Candidate]):
        """
        Add a collection of candiates to the multidispatch. If the multidispatch has no set name or doc, the name or doc of the first candidate with the relevant attributes will be used.

        :param candidates: an iterable of candidates to be added.
        """
        clean_sizes = set()
        for cand in candidates:
            self._add_candidate(cand, clean_cache=False)
            clean_sizes.add(len(cand.types))
        self._clean_cache(clean_sizes)
        return self

    def add_func(self, priority=0, symmetric=False, func=None):
        """
        Adds candidates to a multidispatch generated from a function, usable as a decorator

        :param priority: the priority of the candidates.
        :param symmetric: if set to true, the permutations of all the candidates are added as well
        :param func: the function to used
        """
        if not func:
            if callable(priority):
                func = priority
                priority = 0
            else:
                return partial(self.add_func, priority, symmetric)
        cands = Candidate.from_func(priority, func)
        if symmetric:
            cands = chain.from_iterable(c.permutations() for c in cands)
        self.add_candidates(cands)
        return self

    def _yield_layers(self, types):
        """
        yield all the relevant candidates for a type tuple, sorted first by number of upcasts required (ascending),
        and second by priority (descending)

        :param types: the type tuple to get candidates for
        """
        sub_cache = self.cache.get(len(types))
        if not sub_cache:
            sub_cache = self.cache[len(types)] = {}
            cache = sub_cache[types] = CachedSearch(self, types)
        else:
            cache = sub_cache.get(types)
            if not cache:
                cache = sub_cache[types] = CachedSearch(self, types)

        return cache

    def get(self, args, kwargs, default=None):
        """
        call the multidispatch with args as arguments, attempts all the appropriate candidate until one returns a
        non-NotImplemted value. If all the candidates are exhausted, returns default.

        :param args: the arguments for the multidispatch
        :param kwargs: keyword arguments forwarded directly to any attempted candidate
        :param default: the value to return if all candidates are exhausted
        """
        types = tuple(type(a) for a in args)
        for layer in self.candidates_for_types(*types):
            if len(layer) == 0:
                continue
            if len(layer) != 1:
                raise AmbiguityError(layer, types)
            c = layer[0]
            ret = c.func(*args, **kwargs)
            if ret is not NotImplemented:
                return RawReturnValue.unwrap(ret)
        return default

    def __call__(self, *args, **kwargs):
        """
        call the multidispatch and raise an error if no candidates are found
        """
        ret = self.get(args, kwargs, default=EMPTY)
        if ret is EMPTY:
            raise NoCandidateError(args)
        return ret

    def op(self):
        """
        :return: an adapter for the multidispatch to be used as an adapter, returning NotImplemented if no candidates match,
         and setting the multidispatch's name if necessary
        """
        return MultiDispatchOp(self)

    def method(self):
        """
        :return: an adapter for the multidispatch to be used as a method, raising error if no candidates match,
         and setting the multidispatch's name if necessary
        """
        return MultiDispatchMethod(self)

    def classmethod(self):
        """
        :return: an adapter for the multidispatch to be used as a class method, raising error if no candidates match,
         and setting the multidispatch's name if necessary
        """
        return MultiDispatchClassMethod(self)

    def staticmethod(self):
        """
        :return: an adapter for the multidispatch to be used as a static method, raising error if no candidates match,
         and setting the multidispatch's name if necessary
        """
        return MultiDispatchStaticMethod(self)

    def implementor(self, *args, **kwargs) -> Union[Callable[[Callable], 'Implementor'], 'Implementor']:
        """
        create an Implementor for the MultiDispatch and call its implementor method with the arguments
        """
        return Implementor(self).implementor(*args, **kwargs)

    def candidates_for_types(self, *arg_types) -> Iterator[List[Candidate]]:
        """
        get candidate layers as they are attempted for a set of argument types
        """
        return iter(self._yield_layers(arg_types))

    def candidates(self) -> Iterator[Candidate]:
        """
        get all the candidates defined in the multidispatch.
         Candidates are sorted by their priority, then topologically.
        """
        return map(lambda x: x.inner, self.candidate_set)

    def __str__(self):
        if self.__name__:
            return f'<MultiDispatch {self.__name__}>'
        return super().__str__()



from __future__ import annotations

from typing import TypeVar, Generic, Iterable, MutableMapping, Callable, List, Any

K = TypeVar('K')
V = TypeVar('V')

_blank = object()
_no_default = object()

# todo make index-invariant queries?

class TrieNode(Generic[K, V]):
    """
    A node for a trie
    """
    __slots__ = 'children', 'inner_value'

    def __init__(self):
        self.children: MutableMapping[K, TrieNode[K, V]] = {}
        self.inner_value: V = _blank

    def value(self, default=_no_default):
        """
        :return: the inner value of the node, or default if none exists
        """
        if self.inner_value is _blank:
            if default is _no_default:
                raise ValueError('no value')
            return default
        return self.inner_value

    @property
    def has_value(self):
        """
        :return: whether the node has an inner value
        """
        return self.inner_value is not _blank

    def values(self, buffer: list):
        """
        iterate over the values of the node and its children
        :param buffer: a key buffer to store the sub-keys visited up to the current value
        """
        if self.has_value:
            yield self.value()
        for k, v in self.children.items():
            buffer.append(k)
            yield from v.values(buffer)
            buffer.pop()

    def clear(self):
        self.children.clear()
        self.inner_value = _blank

    def ensure_child(self, sub_key: K) -> TrieNode[K, V]:
        """
        make sure a node has a child of the sub-key, and return it
        """
        ret = self.children.get(sub_key)
        if not ret:
            ret = self.children[sub_key] = TrieNode()
        return ret


class Trie(Generic[K, V], MutableMapping[Iterable[K], V]):
    """
    A Trie that can iterate over nodes
    """

    def __init__(self, update_arg=None, **kwargs):
        self.root: TrieNode[K, V] = TrieNode()
        self._len = 0
        if update_arg or kwargs:
            self.update(update_arg, **kwargs)

    def get(self, key: Iterable[K], default: V = None) -> V:
        current = self.root
        for k in key:
            current = current.children.get(k)
            if not current:
                return default
        ret = current.inner_value
        if ret is _blank:
            return default
        return ret

    def __getitem__(self, item):
        ret = self.get(item, _blank)
        if ret is _blank:
            raise KeyError(item)
        return ret

    def _set(self, key, value, override):
        current = self.root
        for k in key:
            current = current.ensure_child(k)
        current_value = current.inner_value
        if current_value is _blank:
            self._len += 1
        elif not override:
            return current_value
        current.inner_value = value
        return value

    def __setitem__(self, key, value):
        return self._set(key, value, True)

    def setdefault(self, key, default=None):
        return self._set(key, default, False)

    def pop(self, key, default: V = _no_default):
        stack = [(None, self.root)]
        for k in key:
            current = stack[-1][-1].children.get(k)
            if not current:
                break
            stack.append((k, current))
        else:
            current = stack[-1][-1]
            ret = current.inner_value
            if ret is not _blank:
                current.inner_value = _blank
                self._len -= 1
                clear_flag = False
                clear_flag_key = None
                for k, s in reversed(stack):
                    if clear_flag:
                        del s.children[clear_flag_key]
                        clear_flag = False

                    if s.inner_value is _blank and not s.children:
                        clear_flag = True
                        clear_flag_key = k
                return ret

        if default is _no_default:
            raise KeyError(key)

        return default

    def __delitem__(self, key):
        return self.pop(key)

    def _items(self):
        """
        iterate over the key-value pairs in the trie
        note: the keys returned are all returned using the same list buffer, they must be copied or
         converted before returning to the user
        """
        buffer = []
        for v in self.root.values(buffer):
            yield buffer, v

    def __len__(self):
        return self._len

    def __iter__(self, key_converter=tuple):
        return (k for (k, _) in self.items(key_converter))

    def items(self, key_converter: Callable[[List[K]], Any] = tuple):
        if key_converter:
            return ((key_converter(k), v) for (k, v) in self._items())
        return ((k, v) for (k, v) in self._items())

    def values(self):
        return self.root.values([])

    def clear(self):
        self.root.clear()
        self._len = 0



from dyndis.multidispatch import MultiDispatch, RawNotImplemented
from dyndis.type_keys.type_key import Self
from dyndis.type_keys.delegates import UnboundAttr, UnboundMap
from dyndis.type_keys.unbound_friend import UnboundFriend
from dyndis_data import __version__, __author__, __url__



from abc import abstractmethod
from itertools import chain, product
from typing import TypeVar, Dict, MutableSet, Callable

from dyndis.type_keys.type_key import CoreTypeKey, ClassKey, type_key, within_constraints


class UnboundDelegate(CoreTypeKey):
    """
    A delegate for an assigned TypeVar
    """

    class Nil:
        """
        A class that cannot be instantiated, useful for silently failing an unbound delegate
        """

        def __new__(cls):
            raise TypeError('cannot instantiate instance of class Nil')

        def __init_subclass__(cls, **kwargs):
            raise TypeError('cannot subclass Nil')

    def __init__(self, *type_vars: TypeVar):
        """
        :param type_var: the type variable to use
        """
        self.type_vars = type_vars
        self._possibilities = ...

    @abstractmethod
    def __call__(self, *bound_types: type) -> type:
        """
        resolve the delegate

        :param bound_type: the assigned type of the type variable
        :return: the type to delegate to
        """
        return self.Nil

    def get(self, defined_type_var: Dict[TypeVar, ClassKey]):
        defs = [defined_type_var.get(tv) for tv in self.type_vars]
        if None in defs:
            raise UnboundTypeVarError(self, [tv for (tv, b) in zip(self.type_vars, defs) if b is None])
        return type_key(
            self(*(d.inner for d in defs))
        )

    def match(self, query_key: type, defined_type_var: Dict[TypeVar, ClassKey]):
        return self.get(defined_type_var).match(query_key, defined_type_var)

    def possibilities(self):
        """
        :return: all possible values of the delegate, or None if it cannot be calculated
        Node: can only be calculated if all typevars are constrained
        """
        if self._possibilities is ...:
            constraints = []
            for t in self.type_vars:
                if not t.__constraints__:
                    self._possibilities = None
                    break
                constraints.append(t.__constraints__)
            else:
                self._possibilities = tuple(
                    self(*args) for args in product(*constraints)
                )
        return self._possibilities

    def __lt__(self, other):
        pos = self.possibilities()
        if pos is None:
            return NotImplemented
        return within_constraints(pos, other, True)

    def __le__(self, other):
        if self == other:
            return True
        pos = self.possibilities()
        if pos is None:
            return NotImplemented
        return within_constraints(pos, other, False)

    def introduce(self, encountered_type_variables: MutableSet[TypeVar]):
        super().introduce(encountered_type_variables)
        missing = [tv for tv in self.type_vars if tv not in encountered_type_variables]
        if missing:
            raise OrphanDelegateError(self, missing)


_raise = object()


class UnboundAttr(UnboundDelegate):
    """
    A delegate that gets an attribute of a type
    """

    def __init__(self, type_var: TypeVar, attribute: str, default=_raise):
        """
        :param attribute: the attribute name to extract
        :param default: the default value to return if the attribute does not exist, if none is entered, and exception
         will be raised instead.
        """
        super().__init__(type_var)
        self.attribute = attribute
        self.default = default

    def __call__(self, bound_type: type):
        ret = getattr(bound_type, self.attribute, self.default)
        if ret is _raise:
            raise AttributeError(self.attribute)
        return ret

    def __eq__(self, other):
        return type(self) == type(other) and \
               (self.type_vars, self.attribute, self.default) == (other.type_var, other.attribute, other.default)

    def __hash__(self):
        return hash((self.type_vars, self.attribute, self.default))

    def __repr__(self):
        if self.default is _raise:
            return f'{type(self).__name__}({self.type_vars[0].__name__}, {self.attribute!r})'
        return f'{type(self).__name__}({self.type_vars[0].__name__}, {self.attribute!r}, {self.default!r})'


class UnboundMap(UnboundDelegate):
    """
    A delegate that calls a function
    """

    def __init__(self, func: Callable[..., type], *type_vars):
        super().__init__(type_vars)
        self.func = func

    def __call__(self, *bound_types: type):
        return self.func(*bound_types)

    def __eq__(self, other):
        return type(self) == type(other) and \
               (self.type_vars, self.func) == (other.type_var, other.func)

    def __hash__(self):
        return hash((self.type_vars, self.func))

    def __repr__(self):
        return f'{type(self).__name__}(' \
               + ", ".join(chain(
            (repr(self.func),),
            *(t.__name___ for t in self.type_vars)
        )) \
               + ')'


class UnboundTypeVarError(RuntimeError):
    """An error indicating that an UnboundDelegate resolution was attempted when the original type var was not
     assigned, in theory this exception should never be raised, since OrphanDelegateError should be raised at
      function definition."""

    def __init__(self, unbound_delegate, unbound_vars):
        super().__init__(f'type variable {unbound_vars} must be bound before {unbound_delegate}')


class OrphanDelegateError(TypeError):
    """An error indicating that an UnboundDelegate was used as a type hint before its associated typevar"""

    def __init__(self, unbound_delegate, not_defined):
        super().__init__(f'type variable {not_defined} must be used as a type hint before {unbound_delegate}')



class AmbiguityError(RuntimeError):
    """An error indicating that a multidispatch had to decide between candidates of equal precedence"""

    def __init__(self, candidates, types):
        cands = "[" + ", ".join(str(c) for c in candidates) + "]"
        super().__init__(
            f'multiple candidates of equal precedence: {cands} for key <' + ", ".join(t.__name__ for t in types) + ">")


class NoCandidateError(TypeError):
    """An error indicating that a multidispatch has no applicable candidates, or that all candidates failed"""

    def __init__(self, args):
        super().__init__('no valid candidates for argument types <' + ", ".join(type(a).__name__ for a in args) + '>')



from functools import partial
from itertools import chain
from numbers import Number
from typing import NamedTuple, Callable, List, Union
from warnings import warn

from dyndis.candidate import Candidate
from dyndis.type_keys.type_key import Self


class QueuedImplementation(NamedTuple):
    """
    An implementation inside a class that is waiting for the class to complete its creation
    """
    priority: Number
    func: Callable
    permutations: bool

    def cands(self, owner):
        """
        create a set of candidates when the class is complete
        """
        ret = Candidate.from_func(self.priority, self.func, fallback_type_hint=Self, self_type=owner)
        if self.permutations:
            ret = chain.from_iterable(r.permutations() for r in ret)
        return ret


class Implementor:
    """
    A descriptor that holds candidates that include a class until it is complete
    """

    def __init__(self, multidispatch):
        """
        :param multidispatch: the multidispatch to add the candidates to
        """
        self.multidispatch = multidispatch

        self.queue: List[QueuedImplementation] = []
        self.locked = False

    def __set_name__(self, owner, name):
        """
        indicates that the class is fully created, adds the queued implementors to the multidispatch
        """

        def check_candidate(candidate):
            if self.multidispatch.__name__ is not None \
                    and candidate.__name__ is not None \
                    and self.multidispatch.__name__.strip('_') != candidate.__name__.strip('_'):
                warn(f'implementor {candidate} in class {owner.__name__} has dissimilar name from multidispatch'
                     f' {self.multidispatch}', stacklevel=5)
            return candidate

        if self.locked:
            return
        self.multidispatch.add_candidates(
            check_candidate(c) for c in
            chain.from_iterable(q.cands(owner) for q in self.queue)
        )
        self.locked = True
        # self-destruct
        delattr(owner, name)

    def implementor(self, priority=0, symmetric=False, func=None) \
            -> Union[Callable[..., 'Implementor'], 'Implementor']:
        """
        Add a new queued candidate, usable as a decorator
        :param priority: the priority of the candidate
        :param symmetric: if set to true, all the permutations of the candidate will also be added
        :param func: the function to be added
        """
        if not func:
            return partial(self.implementor, priority, symmetric)

        if self.locked:
            raise Exception('the implementor has already been locked, no new functions can be added')
        self.queue.append(QueuedImplementation(priority=priority, func=func, permutations=symmetric))
        return self



from __future__ import annotations

from functools import lru_cache
from typing import Union, Dict, TypeVar, Hashable, Any, Iterable, Optional, Tuple, ByteString, Generic, MutableSet
from abc import abstractmethod, ABC

from dyndis.util import get_args, get_origin, liberal_cache

try:
    from typing import Literal, Protocol
except ImportError:
    Literal = None
    Protocol = None


class MatchException(RuntimeError):
    """
    An error that occurs during a match, and might be delayed until its rank is reached
    """


class AmbiguousBindingError(MatchException):
    """An error indicating that a type variable could not find a single type to bind to"""

    def __init__(self, typevar, subclass, unrelated_classes):
        super().__init__(f'type variable {typevar} must up-cast type {subclass} to one of its constrained types,'
                         f' but it is a subclass of multiple non-related constraints: {unrelated_classes}'
                         f' (consider adding {subclass} as an explicit constraint in {typevar},'
                         f' or a specialized overload for {subclass})')


class TypeKey(ABC):
    """
    A key to be used to identify a parameter
    """

    @abstractmethod
    def match(self, query_key: type, defined_type_var: Dict[TypeVar, ClassKey]) \
            -> Union[bool, MatchException]:
        """
        Evaluate the query key into the type key
        :param query_key: the type of the parameter to be matched
        :param defined_type_var: a dictionary mapping already bound type variables to their associated types
        :return: Whether there is a match, or a MatchException if an error occurred
        """
        pass

    @abstractmethod
    def __le__(self, other: CoreTypeKey):
        """
        A <= B if for every class c that A matches, B also matches c
        """
        pass

    @abstractmethod
    def __lt__(self, other: CoreTypeKey):
        """
        A < B if A <= B and also there exists class c that B matches s.t. A doesn't match c
        """
        pass

    def __call__(self):
        # typing's structures (like Union) will reject all non-callable arguments,
        # so we have to make TypeKey seem callable
        raise TypeError(f"can't call {type(self)}")


@liberal_cache
def type_key(t) -> TypeKey:
    """
    convert type annotation in a (possibly non-core) type key
    """

    def is_type_like(ga):
        # the best way I know of to check whether a generic alias is usable as class
        try:
            isinstance(None, ga)
        except TypeError:
            return False
        return True

    if isinstance(t, TypeKey):
        return t
    if isinstance(t, type):
        if t in SplittingClassTypeKey.TYPE_ALIASES:
            return SplittingClassTypeKey(t)
        return ClassKey(t)
    if isinstance(t, TypeVar):
        return TypeVarKey(t)
    if t is Any:
        return AnyKey
    if t in (..., NotImplemented, None):
        return SingletonTypeKey(t)

    origin = get_origin(t)

    if Literal and origin is Literal:
        return LiteralTypeKey(t)
    if origin is Union:
        return UnionTypeKey(t)
    if isinstance(origin, type) and is_type_like(t):
        return type_key(origin)
    raise TypeError(f'type annotation {t} is not a type, give it a default to ignore it from the candidate list')


def class_type_key(t) -> ClassKey:
    """
    :return: convert the argument into a ClassTypeKey, useful for when we would like to avoid splitting type keys
    """
    tk = type_key(t)
    # special case for splitting class types
    if isinstance(tk, SplittingClassTypeKey):
        tk = next(tk.split()),
    else:
        tk = type_keys(tk)
    if len(tk) != 1 or not isinstance(tk[0], ClassKey):
        raise Exception(f'type annotation {t} must evaluate to a single type')

    return tk[0]


@liberal_cache
def type_keys(t) -> Tuple[Union[CoreTypeKey, SelfKeyCls]]:
    """
    Convert a type hint to a tuple of core type keys (or SelfKey)
    :param t: the type hint
    :return: a tuple of core or self type keys
    can handle:
    * types
    * singletons (..., None, Notimplemented)
    * the typing.Any object
    * the dyndis.Self object
    * any non-specific typing abstract class (Sized, Iterable, ect...)
    * Type variables
    * typing.Union
    * dyndis.UnboundDelegate object
    * Any TypeKey object
    3.8 only:
    * Literals of singleton types
    """

    def recursive_split(tk: TypeKey) -> Iterable[Union[CoreTypeKey, SelfKeyCls]]:
        if isinstance(tk, SplittingTypeKey):
            for t in tk.split():
                yield from recursive_split(t)
        else:
            yield tk

    tk = type_key(t)
    return tuple(recursive_split(tk))


T = TypeVar('T', bound=Hashable)


class WrapperKey(TypeKey, Generic[T], ABC):
    """
    A mixin class for a type key that simply wraps a another value
    """

    def __init__(self, inner: T):
        self.inner = inner


# region core
class CoreTypeKey(TypeKey):
    """
    A core type key is one that be hashed into a dict (specifically a trie), it usually has the same hash as a
     wrapped value
    """

    def introduce(self, encountered_type_variables: MutableSet[TypeVar]):
        """
        called when the type_key is being used to construct a candidate
        :param encountered_type_variables: the type variables previously encountered in previous parameters
        """
        pass

    @abstractmethod
    def __repr__(self):
        """
        Core type keys appear in candidate representations, so they must look like the annotation used ot create them
        """
        pass


class CoreWrapperKey(WrapperKey[T], CoreTypeKey, Generic[T], ABC):
    """
    A mixin type key class or core and wrapper
    """

    def __repr__(self):
        return repr(self.inner)


def within_constraints(constraints, other, strict):
    """
    :param constraints: the constraints to compare
    :param other: the other element to use
    :param strict: if set to True, will compare for strict matches, otherwise, will compare for equal-or-less
    :return: whether all constraints are either less than or not comparable
    """
    ret = NotImplemented
    for c in constraints:
        try:
            ctk = type_key(c)
            cmp = ctk < other if strict else ctk <= other
            rev_cmp = other <= other if strict else other < ctk
        except TypeError:
            continue
        if cmp:
            if ret is False:
                return NotImplemented
            ret = True
        elif rev_cmp:
            if ret is True:
                return NotImplemented
            ret = False
    return ret


class ClassKey(CoreWrapperKey[type]):
    """
    A type key of a superclass
    """

    def __init__(self, *args):
        super().__init__(*args)
        assert isinstance(self.inner, type)

    def match(self, query_key: type, defined_type_var) -> Union[bool, Exception]:
        return query_key is self.inner or issubclass(query_key, self.inner)

    def __le__(self, other):
        if isinstance(other, ClassKey):
            if issubclass(self.inner, other.inner):
                return True
            if issubclass(other.inner, self.inner):
                return False
            return NotImplemented
        try:
            rev_cmp = other < self
        except TypeError:
            return NotImplemented
        return not rev_cmp

    def __lt__(self, other):
        if isinstance(other, ClassKey):
            if self.inner == other.inner or issubclass(other.inner, self.inner):
                return False
            if issubclass(self.inner, other.inner):
                return True
            return NotImplemented
        try:
            rev_cmp = other < self
        except TypeError:
            return NotImplemented
        return not rev_cmp

    def __repr__(self):
        return self.inner.__name__


class TypeVarKey(CoreWrapperKey[TypeVar]):
    """
    A type key of a type variable
    """

    @staticmethod
    @lru_cache()
    def constrain_type(cls, scls: Union[type, TypeVar]) -> Optional[ClassKey]:
        """
        get the lowest type that cls can be up-cast to and scls accepts as constraint. Or None if none exists.
        """
        if isinstance(scls, TypeVar):
            if scls.__constraints__:
                candidates = [c for c in scls.__constraints__ if issubclass(cls, c)]
                if not candidates:
                    return None
                minimal_candidates = [
                    cand for cand in candidates if all(issubclass(cand, c) for c in candidates)
                ]
                if len(minimal_candidates) != 1:
                    raise AmbiguousBindingError(scls, cls, minimal_candidates or candidates)
                return class_type_key(minimal_candidates[0])
            elif scls.__bound__:
                return TypeVarKey.constrain_type(cls, scls.__bound__)
            return class_type_key(cls)
        return class_type_key(cls) if issubclass(cls, scls) else None

    def match(self, query_key: type, defined_type_var: Dict[TypeVar, ClassKey]) -> Union[bool, MatchException]:
        return defined_type_var[self.inner].match(query_key, defined_type_var)

    def __le__(self, other):
        if self == other:
            return True

        if self.inner.__constraints__:
            return within_constraints(self.inner.__constraints__, other, strict=True)
        if isinstance(other, ClassKey):
            return class_type_key(self.inner.__bound__ or object) <= other
        return NotImplemented

    def __lt__(self, other):
        if self == other:
            return False

        if self.inner.__constraints__:
            return within_constraints(self.inner.__constraints__, other, strict=True)
        if isinstance(other, ClassKey):
            return class_type_key(self.inner.__bound__ or object) <= other
        return NotImplemented

    def introduce(self, encountered_type_variables: MutableSet[TypeVar]):
        super().introduce(encountered_type_variables)
        encountered_type_variables.add(self.inner)


class AnyKeyCls(CoreWrapperKey[type(Any)]):
    """
    A singleton type key for the Any object, that encompasses all other keys
    """
    def __init__(self):
        super().__init__(Any)

    def match(self, query_key: type, defined_type_var: Dict[TypeVar, ClassKey]) -> Union[bool, Exception]:
        return True

    def __le__(self, other):
        return other is self

    def __lt__(self, other):
        return False

    def __ge__(self, other):
        return True

    def __gt__(self, other):
        return other is not self


AnyKey = AnyKeyCls()


# endregion


# region splitters
class SplittingTypeKey(TypeKey):
    """
    A splitting type key is one that can be split into core type keys. splitting type keys are
     unhashable to ensure they don't end up in the candidate structure.
    """

    @abstractmethod
    def split(self) -> Iterable[Union[SplittingTypeKey, CoreTypeKey]]:
        """
        split the type key into smaller (possibly core) type keys
        """
        pass

    def match(self, query_key: type, defined_type_var: Dict[TypeVar, ClassKey]) \
            -> Union[bool, MatchException]:
        for s in self.split():
            r = s.match(query_key, defined_type_var)
            if isinstance(r, MatchException):
                return r
            if r:
                return r
        return False

    def __lt__(self, other):
        split = list(self.split())
        if within_constraints(split, other, True):
            return True
        if any(other <= s for s in split):
            return False
        return NotImplemented

    def __le__(self, other):
        split = self.split()
        if within_constraints(split, other, False):
            return True
        if any(other < s for s in split):
            return False
        return NotImplemented

    __hash__ = None


class SplittingClassTypeKey(WrapperKey[type], SplittingTypeKey):
    """
    A class type key for classes that, according to various PEPs, should be treated as unions
    """
    TYPE_ALIASES: Dict[type, Tuple[type, ...]] = {
        float: (float, int),
        complex: (float, int, complex),
        bytes: (ByteString.__origin__,)
    }

    def __init__(self, inner):
        super().__init__(inner)
        self.aliases = self.TYPE_ALIASES[inner]

    def split(self) -> Optional[Iterable[TypeKey]]:
        return (ClassKey(a) for a in self.aliases)


class SingletonTypeKey(WrapperKey[Union[None, type(NotImplemented), type(...)]], SplittingTypeKey):
    """
    A type key for singleton types
    """

    def __init__(self, inner):
        if inner not in (..., NotImplemented, None):
            raise TypeError('cannot have non-singleton in singleton key')
        super().__init__(inner)

    def split(self) -> Optional[Iterable[TypeKey]]:
        return type_key(type(self.inner)),


class UnionTypeKey(SplittingTypeKey, WrapperKey):
    """
    A type key for union types
    """

    def split(self) -> Iterable[TypeKey]:
        return (type_key(a) for a in get_args(self.inner))


if Literal:
    class LiteralTypeKey(WrapperKey, SplittingTypeKey):
        """
        A type key for literal types
        """

        def __init__(self, inner):
            if frozenset(get_args(inner)) <= frozenset((..., NotImplemented, None)):
                raise TypeError('cannot have non-singleton in literal key')
            super().__init__(inner)

        def split(self) -> Iterable[TypeKey]:
            return (type_key(a) for a in get_args(self.inner))


# endregion

class SelfKeyCls(TypeKey):
    """
    A special type key that evaluates to the candidate's self_type when the candidate is created
    """

    def raise_(self, *args, **kwargs):
        raise TypeError('Self is a special type kay that must not actually be used')

    match = __le__ = __lt__ = raise_


Self = SelfKeyCls()



from __future__ import annotations

from itertools import chain
from typing import TypeVar, Set, Union, MutableSet, Any, MutableMapping, List, Iterator, Sequence

from sortedcontainers import SortedDict

from dyndis.candidate import Candidate
from dyndis.trie import Trie, TrieNode
from dyndis.type_keys.type_key import TypeKey, MatchException, TypeVarKey

T = TypeVar('T')


class TopologicalNode:
    """
    A node of an element in a topological set
    """
    __slots__ = 'inner', 'gt', 'lt', '_layer'

    def __init__(self, inner: Candidate):
        self.inner = inner
        # rule: if root is in lt, then len(gt) == 1
        self.lt: Set[Union[TopologicalNode, TopologicalRootNode]] = set()
        self.gt: Set[TopologicalNode] = set()

        self._layer = None

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.inner < other.inner

    def layer(self):
        if self._layer is None:
            self._layer = max(l.layer() for l in self.lt) + 1
        return self._layer

    def invalidate_layer(self):
        self._layer = None


class TopologicalRootNode:
    """
    The virtual root node of a topological set. This is a handy wrapper for a set's first layer, allowing it to be used
     as a node
    """

    def __init__(self, owner: TopologicalSet):
        self.owner = owner

    @property
    def gt(self):
        if not self.owner.layers:
            return frozenset()
        return self.owner.layers[0]

    def __lt__(self, other):
        return True

    def layer(self):
        return -1


class CandidateSet(MutableSet[TopologicalNode]):
    """
    A set that stores candidates of equal rank in easy-to-query tries
    """

    def __init__(self):
        self.tries: MutableMapping[Any, Trie[TypeKey, TopologicalNode]] = SortedDict()

    def add(self, cand: TopologicalNode):
        trie = self.tries.get(cand.inner.priority)
        if trie is None:
            trie = self.tries[cand.inner.priority] = Trie()
        trie[cand.inner.types] = cand

    def remove(self, cand: TopologicalNode):
        trie = self.tries.get(cand.inner.priority)
        if trie is None:
            raise KeyError(cand)
        del trie[cand.inner.types]
        if not trie:
            del self.tries[cand.inner.priority]

    def discard(self, x) -> None:
        try:
            self.remove(x)
        except KeyError:
            pass

    def __len__(self):
        return sum(len(t) for t in self.tries.values())

    def __iter__(self):
        for trie in reversed(self.tries.values()):
            yield from trie.values()

    def clear(self) -> None:
        self.tries.clear()

    def __contains__(self, cand):
        trie = self.tries.get(cand.inner.priority)
        if trie is None:
            return False
        return cand.inner.types in trie

    def by_sublayer(self, query: Sequence[type]):
        """
        :return: for each priority in self (in decreasing order), yields all the candidates of that priority that match
         the query
        """

        def _matches(node: TrieNode[TypeKey, TopologicalNode], ind, defined_type_vars):
            if ind == len(query):
                if node.has_value:
                    yield node.value()
                return
            q = query[ind]
            for k, child in node.children.items():
                new_dtv = defined_type_vars
                if isinstance(k, TypeVarKey) and k.inner not in new_dtv:
                    constrained = k.constrain_type(q, k.inner)
                    if not constrained:
                        continue
                    new_dtv = dict(new_dtv)
                    new_dtv[k.inner] = constrained
                    match = True
                else:
                    match = k.match(q, new_dtv)
                    if isinstance(match, MatchException):
                        raise match

                if match:
                    yield from _matches(child, ind + 1, new_dtv)

        for t in reversed(self.tries.values()):
            yield _matches(t.root, 0, {})


class TopologicalSet(MutableSet[Candidate]):
    """
    A set that allows for quick topological sorting of its elements, supporting weakly ordered elements
    """

    def __init__(self, arg=()):
        self.root = TopologicalRootNode(self)
        self.layers: List[CandidateSet] = []
        self._len = 0
        for a in arg:
            self.add(a)

    def add(self, x) -> bool:
        """
        :return: True if all went well, False if an equivalent element was found in the set
        """
        node = TopologicalNode(x)

        potential_parents: Set = {self.root}
        direct_parents = []
        direct_children = []
        while potential_parents:
            pp = potential_parents.pop()
            any_subchildren = False
            for k in pp.gt:
                if k < node:
                    potential_parents.add(k)
                    any_subchildren = True
                elif node < k:
                    direct_children.append(k)
                elif node.inner <= k.inner and k.inner >= node.inner:
                    return False
            if not any_subchildren:
                direct_parents.append(pp)

        for dp in direct_parents:
            node.lt.add(dp)
            if dp is not self.root:
                dp.gt.add(node)

        for dc in direct_children:
            supplanted = dc.lt & node.lt
            for s in supplanted:
                if s is not self.root:
                    s.gt.remove(dc)
            dc.lt.difference_update(supplanted)
            dc.lt.add(node)
            node.gt.add(dc)
            self._rebase(dc)

        layer = node.layer()
        if len(self.layers) == layer:
            self.layers.append(CandidateSet())
        self.layers[layer].add(node)

        return True

    def _rebase(self, candidate: TopologicalNode):
        """
        remove a node from its current layer, and place it in its appropriate layer, and similarly rebase all its
         children
        """
        prev_layer = candidate.layer()
        candidate.invalidate_layer()
        new_layer = candidate.layer()
        if new_layer == prev_layer:
            return
        self.layers[prev_layer].remove(candidate)
        if len(self.layers) <= new_layer:
            self.layers.append(CandidateSet())
        self.layers[new_layer].add(candidate)
        for child in candidate.gt:
            self._rebase(child)

    def __contains__(self, item):
        raise NotImplementedError

    def remove(self, x) -> None:
        raise NotImplementedError

    def discard(self, x) -> None:
        try:
            self.remove(x)
        except KeyError:
            pass

    def __iter__(self):
        return chain.from_iterable(self.layers)

    def __len__(self):
        return self._len

    def clear(self) -> None:
        self._len = 0
        self.root.gt.clear()
        self.layers.clear()



from typing import Any, NamedTuple

try:
    # get_origin and get_args are only defined in 3.8
    from typing import get_origin, get_args
except ImportError:
    def get_origin(tp):
        return getattr(tp, '__origin__', None)


    def get_args(tp):
        return getattr(tp, '__args__', ())


class RawReturnValue(NamedTuple):
    """
    A class to wrap otherwise special return values from a multidispatch candidate
    """
    inner: Any

    @classmethod
    def unwrap(cls, x):
        """
        If x is a RawReturnValue, return its inner value, if not, return x unchanged
        """
        if isinstance(x, cls):
            return x.inner
        return x


class SubPriority:
    """
    A generic priority with a secondary modifier
    """

    @classmethod
    def make(cls, x, weight=-1):
        """
        create a sub-priority
        """
        if weight == 0:
            return x
        return cls(x, weight)

    def __init__(self, original, weight):
        self.original = original
        self.weight = weight
        self.key = (self.original, self.weight)

    @classmethod
    def to_key(cls, x):
        """
        force x into a key that can be compared with self.key
        """
        if isinstance(x, cls):
            return x.key
        return x, 0

    def __lt__(self, other):
        return self.key < self.to_key(other)

    def __le__(self, other):
        return self.key <= self.to_key(other)

    def __gt__(self, other):
        return self.key > self.to_key(other)

    def __ge__(self, other):
        return self.key >= self.to_key(other)

    def __eq__(self, other):
        return self.key == self.to_key(other)

    def __hash__(self):
        return hash(self.key)


def liberal_cache(func):
    """
    create a memoized function that simply ignores un-hashable values
    """
    cache = {}

    def ret(arg):
        try:
            return cache[arg]
        except TypeError:
            return func(arg)
        except KeyError:
            pass
        res = cache[arg] = func(arg)
        return res

    return ret



from inspect import isclass, ismodule
from itertools import chain
from types import SimpleNamespace, MappingProxyType
from typing import Callable
from warnings import warn

import sys

from dyndis.type_keys.delegates import UnboundDelegate, _raise


class UnboundFriend(UnboundDelegate):
    """
    Searches for a class with a similar name in the type's namespace
    """

    def __init__(self, name_func: Callable[..., str], *type_vars, namespace=..., default=_raise):
        """
        :param name_func: a callable to return the name of the attribute to search for
        :param namespace: the name space to search in, default will search in the namespaces of the types
        :param default: if provided, this value will be returned instead of raising an error
        """
        super().__init__(type_vars)
        self.name_func = name_func
        self.namespace = namespace
        self.default = default

    def __call__(self, *bound_types: type):
        name = self.name_func(*bound_types)
        if self.namespace is ...:
            mods = []
            for bt in bound_types:
                try:
                    mod = get_namespaces(bt)
                except (TypeError, NameError) as e:
                    warn(f'error while extracting namespace of {bt}: {e}')
                else:
                    mods.append(mod)
            ns = JoinedNamespaces.make(mods)
        elif isinstance(self.namespace, str):
            ns = sys.modules[self.namespace]
        elif isinstance(self.namespace, (dict, MappingProxyType)):
            ns = SimpleNamespace()
            ns.__dict__.update(self.namespace)
        else:
            ns = self.namespace

        try:
            return getattr(ns, name)
        except NameError:
            if self.default is _raise:
                raise
            return self.default

    def __eq__(self, other):
        return type(self) == type(other) and \
               (self.type_vars, self.name_func, self.namespace, self.default) \
               == (other.type_var, other.name_func, other.namespace, other.default)

    def __hash__(self):
        return hash((self.type_vars, self.name_func, self.namespace, self.default))

    def __repr__(self):
        return f'{type(self).__name__}(' \
               + ", ".join(chain(
            (repr(self.name_func),),
            *(t.__name___ for t in self.type_vars)
        )) \
               + (", namespace=" + repr(self.namespace) if self.namespace is not ... else "") \
               + (", default=" + repr(self.default) if self.default is not _raise else "") \
               + ')'


def get_namespaces(cls: type):
    """
    get the total namespace where a class is defined
    """
    qname = cls.__qualname__
    mod_name = cls.__module__
    if qname.startswith(mod_name):
        qname = qname[len(mod_name):]
    mod = sys.modules[mod_name]
    ret = [mod]
    for name in qname.split('.'):
        ns = getattr(ret[-1], name, None)
        if ns is None:
            raise NameError(f'cannot reach {cls.__qualname__}')
        if not isclass(ns) and not ismodule(ns):
            raise TypeError(f'{ns} is not a class')
        if ns is cls:
            return StackedNamespaces.make(ret)
        ret.append(ns)
    raise NameError(f'{cls.__qualname__} path does not lead to class)')


class CompoundNamespaces:
    """
    A base class for combining namespaces
    """

    def __init__(self, parts):
        self._parts = parts

    @classmethod
    def make(cls, parts):
        """
        create a compound namespace if necessary
        """
        if len(parts) > 1:
            return cls(parts)
        return parts[0]


class StackedNamespaces(CompoundNamespaces):
    """
    A compound namespace where the last namespace has precedence
    """

    def __init__(self, parts):
        super().__init__(parts[::-1])

    def __getattr__(self, item):
        for p in self._parts:
            try:
                return getattr(p, item)
            except AttributeError:
                pass
        raise NameError(item)


class JoinedNamespaces(CompoundNamespaces):
    """
    A compound namespace where the no one has precedence
    """

    def __getattr__(self, item):
        ret = set()
        for p in self._parts:
            try:
                ret.add(getattr(p, item))
            except AttributeError:
                pass
        if not ret:
            raise NameError(item)
        if len(ret) > 1:
            raise NameError(f'multiple conflicting matches for {item}')
        return next(iter(ret))
